/*
SAS Genetic Marker Analytical Process

Name: Arlequin Input Engine
Author: Xiang Yu, Russ Wolfinger, SAS Institute Inc.
Support: Xiang Yu, Russ Wolfinger, SAS Institute Inc.
History:
   sasrdw 08Dec03 initial coding                          

Description: 
The Arlequin Input Engine imports genetic
marker data in the Arlequin format into a SAS data set and
uploads it to the warehouse.
*/

/*
-----------------------------------------------------------
Copyright (C) 2004 SAS Institute, Inc. All rights reserved. 

Notice: 
The above copyright notice and this notice must appear in
any whole or partial copy of this code and any related
documentation. Permission to use, copy, and modify the
source code contained in this file is hereby granted, but
is limited to customers of SAS with a valid license for the
SAS software product with which this file was provided. 
Except as expressly set forth herein, the terms of such
license apply.
-----------------------------------------------------------
*/

*ProcessVariables;
%let InFile = ; 
* Name of input Arlequin file with full path;

%let OutData = ; 
* Name of SAS data set created from the Arlequin file and uploaded to the warehouse.;

%let OutPath = ; 
* Path for output on your SAS server. If you leave this blank the system 
will use a default location.;

%let UploadInFile = No;
* Specify "Yes" to upload InFile (in original format) into warehouse;

%let MacroPath = ; 
* Path for SAS macros.;



*ProcessBody;

%macro ParseArlequin;
*initialization;
%let ProcessName = ArlequinInputEngine;

%global exiterror;
%let exiterror = 0;
%let clockrunning = 0;
options nospool;
%if &syserr %then %do;
   %put ERROR: Path for utility macros is invalid.;
   %goto exit;
%end;
%include "&MacroPath.UtilityMacros.sas" / nosource;
%ClockStart(&ProcessName);
%if &syserr %then %goto exit;
%let clockrunning = 1;

/* %CheckForBlank(&DataPath, Full path of data);
%if &exiterror %then %goto exit;
%CheckForBlank(&OutPath, Output path);
%if &exiterror %then %goto exit; */

%CheckForBlank(&InFile, arlequin file);
%if &exiterror %then %goto exit; 

%PathName(&InFile);
%if &exiterror %then %do;
   %put ERROR: Arlequin file specification is invalid.;
   %goto exit;
%end;

* create output libaray;
libname OutLib "&OutPath";

%CheckDataName(_name=OutData,_default=&ProcessName);

%include "&MacroPath.UtilityMacros.sas" / nosource;    
%let VeryStart = %sysfunc(round(%sysevalf(%sysfunc(time())),.01));
%let Day1 = %sysfunc(round(%sysevalf(%sysfunc(date()))));

*Parsing an Arlequin file;
/* %let flagdot = 0;
%if (%length(&InFile)>4) & (%upcase(%substr(&InFile,%length(&InFile)-3,4))=.ARP) %then %do;
%let datafile=&DataPath.&InFile;
%let flagdot = 1;
%end; 
%if &flagdot = 0 %then %do; 
	filename DataFil1 "&DataPath.&InFile..ARP";
	filename DataFil2 "&DataPath.&InFile..arp";
%if %sysfunc(fexist(DataFil1)) %then %let Datafile=&DataPath.&InFile..ARP;
%else %if %sysfunc(fexist(DataFil2)) %then %let Datafile=&DataPath.&InFile..arp;
%else %put ERROR: &DataPath.&InFile dose not exist.;
%end; */
%let Datafile = &InFile;

%let FileStartRow = 0;
%let DataBlockRow = 0;
%let DataStartRow = 0;
%let FirstRowOfData = 0; 
%let MissingLabel = .;
%let Separator =;
%let MissingAllele = .;
%let IsHaploid = 0;
%let HapDefRow = 0;
%let InternalHapList = 0;

/* Parse headers */
data _null_;
	infile "&Datafile" TRUNCOVER EXPANDTABS lrecl=32767 firstobs=1; 
	LENGTH var1 $ 256;
	input var1 1-256;
	var1 = upcase(var1);
	if trim(left(var1)) = '[PROFILE]' then call symput('FileStartRow', _n_);
	if index(var1, 'NBSAMPLES=') ^= 0 then 
			do;
				npops = substr(var1, index(var1, 'NBSAMPLES=')+10);				
				call symput('NPops', trim(left(npops)));
			end;
	if index(var1, 'MISSINGDATA=') ^= 0 then 
			do;
				missinglabel = substr(var1, index(var1, 'MISSINGDATA=')+12); 
				missinglabel = translate(missinglabel, '', '"', '', "'");
				call symput('MissingLabel', trim(left(missinglabel)));
			end;
	if index(var1, 'LOCUSSEPARATOR=') ^= 0 then 
			do;
				separator = substr(var1, index(var1, 'LOCUSSEPARATOR=')+15);
				if (trim(left(separator)) = 'WHITESPACE' | trim(left(separator)) = 'TAB') then separator = '';
				call symput('Separator', trim(left(separator)));
			end;
	if index(var1, 'GENOTYPICDATA=') ^= 0 then 
			do;
				genotypeflag = substr(var1, index(var1, 'GENOTYPICDATA=')+14, 1);
				if trim(left(genotypeflag)) = '0' then call symput('IsHaploid', trim(left(1)));
			end;
	if index(var1, '[DATA]') ^= 0 then call symput("DataBlockRow", _n_);
	if index(var1, '[[HAPLOTYPEDEFINITION]]') ^= 0 then call symput('HapDefRow', _n_);		
	if index(var1, '[[SAMPLES]]') ^= 0 then call symput('DataStartRow', _n_);
	if index(var1, 'SAMPLEDATA') ^= 0 then
		do;
			call symput('FirstRowOfData', _n_);
			stop;
		end; 
run;
%if &FileStartRow = 0 %then %put ERROR: Invalid arlequin file, must begin with "[Profile]";
%if &DataBlockRow = 0 %then %put ERROR: Invalid arlequin file, can not find data block starting with command "[Data]";
%if &DataStartRow = 0 %then %put ERROR: Invalid arlequin file, can not find data starting with "[[Samples]]";
%if &FirstRowOfData = 0 %then %put ERROR: Invalid arlequin file, can not find actual data starting with "SampleData=";

*Import external haplotype list;
%if &HapDefRow ^= 0 %then 
	%do;
		data _null_;
			infile "&Datafile" firstobs=%eval(&HapDefRow+1) TRUNCOVER EXPANDTABS lrecl=32767; 
			length v1 $ 512;
			input v1 1-512;
			v1 = upcase(v1);
			if index(v1, 'HAPLLIST=') ^= 0 then
				do;
					if index(v1, 'EXTERN') ^= 0 then
						do;
							haplistfile = substr(v1, index(v1, 'EXTERN')+6);
							haplistfile = translate(haplistfile, '', "'", '', '"');
							call symput('HapListFile', trim(left(haplistfile)));
						end;
					else
						do;
							call symput('InternalHapList', trim(left(1)));
							call symput('HapListRow', _n_);							
						end;
					stop;
				end;
		run;
		%if &InternalHapList = 0 %then 			
			%do;
				data _null_;
					infile "&DataPath.&HapListFile" TRUNCOVER EXPANDTABS lrecl=32767; 
	      			length v1 $1024;
					input v1 1-1024;
					i = 1;
					do while(trim(left(scan(v1, i, " ,;&Separator"))) ^= '');
						i = i+1;
					end;					
					call symput('NumVars', trim(left(i-1)));
					stop;
				run;
				data haplist; 
					infile "&DataPath.&HapListFile" TRUNCOVER EXPANDTABS lrecl=32767;  
					length v1 $ 1024;
					retain len1-len&NumVars 0;
					input v1 1-1024;
					%do i = 1 %to &NumVars;
						var&i = trim(left(scan(v1, &i, " ,;&Separator")));	
						len&i = max(length(var&i), len&i);
					%end;
					if end then 
						do;
							%do k = 1 %to &NumVars;
								call symput("Len&k", len&k);
							%end;
						end;
					id = _n_;
					drop v1 end;
				run;
				proc sort data = haplist out=tmp;
					by descending id;
				run;
				data _null_; set tmp(obs=1);
					%do i = 1 %to &NumVars;
						call symput("Len&i", len&i);
					%end;
				run;
				data haplist; set haplist;
					length hID $ &Len1;
					hID = var1;
					%do i = 2 %to &NumVars;
						%let j = %eval(&i - 1);
						length A&j $ &&Len&i;
						A&j = var&i;
					%end;
					drop id len1-len&NumVars var1-var&NumVars;
				run;
				proc sort data=haplist;
					by hID;
				run;
			%end;
		%else 
			%do;
				data _null_;
					infile "&Datafile" firstobs=%eval(&HapListRow+1) TRUNCOVER EXPANDTABS lrecl=32767; 
					length v1 $ 512;
					input v1 1-512;
					i = 1;
					do while(trim(left(scan(v1, i, " ,;&Separator"))) ^= '');
						i = i+1;
					end;					
					call symput('NumVars', trim(left(i)));
					stop;
				run;
				data _null_; 
					infile "&Datafile" firstobs=%eval(&HapListRow+1) TRUNCOVER EXPANDTABS lrecl=32767; 
					length v1 $ 512;
					do n = 1 to &NumVars;
						retain len&n 0;
					input v1 1-512;
					do i = 1 to &NumVars;
						var&i = trim(left(scan(v1, i, " ,;&Separator")));	
						len&i = max(length(var&i), len&i);
					end;
					if index(v1, '}') ^= 0 then 
						do;
							call symput("EndHapListRow", _n_);
							do k = 1 to &NumVars;
								call symput("Len&k", len&k);
							end;
						end;
				run;
				%let NM = %eval(&NumVars - 1);
				data haplist;
					infile "&Datafile" firstobs=%eval(&HapListRow+1) TRUNCOVER EXPANDTABS lrecl=32767; 
					length hID &Len1;
					do i = 2 to &NumVars;
						j = i - 1;
						length A&j len&i;
					end;
					input hID A1-A&NM;
					if _n_ >= %eval(&EndHapListRow - &HapListRow) then stop;						 
				run;
			%end;	
	%end;

*Check number of markers;
%if &HapDefRow = 0 %then
	%do;
		data _null_;
			infile "&Datafile" firstobs=%eval(&FirstRowOfData+1) TRUNCOVER EXPANDTABS lrecl=32767; 
			length v1 $ 2048;
			input v1 1-2048;
			i = 1;
			do while(trim(left(scan(v1, i, " ,;&Separator"))) ^= '');
				i = i+1;
			end;
			*each row of data has one ID, one frequency number, and a set of markers;
			nloci = i - 3;
			call symput('NLoci', trim(left(nloci)));
			stop;
		run;
	%end;
%else 
	%do;
		*each row of data has one frequency number and one haplotype ID;
		%let NLoci = 1;
	%end;

*Haplotypic data;
%if &IsHaploid = 1 %then
	%do;
		*Using haplotype list;
		%if &HapDefRow ^= 0 %then
			%do;
				*Read in data;
				data arlequin;
					infile "&Datafile" firstobs=%eval(&DataStartRow+1) TRUNCOVER EXPANDTABS lrecl=32767; 
					retain flag 1;
					retain datastart 0; 
		 			retain npopulation 1;
					retain pop;
					length hID $ &Len1;
					length v1 $ 1024;
					input v1 1-1024;			
					if index(v1, 'SampleName') ^= 0 then 
						do;
							pop = substr(v1, index(v1, 'SampleName')+11);
							pop = translate(pop, '', '"', '', "'");	
						end;  			
					if index(v1, 'SampleData') ^= 0 then datastart = 1;
					if datastart = 1 & index(v1, 'SampleData') = 0 then
						do;			
							hID = scan(v1, 1, " ,;&Separator");
							freq = scan(v1, 2, " ,;&Separator") + 0;
							flag = 0;	
						end;
					if trim(left(v1)) = '' then flag = 1;
					if index(v1, '}') ^= 0 then 
						do;
							datastart = 0;
							npopulation = npopulation + 1;
							flag = 1;
						end;
					if npopulation > &NPops then stop;
					drop v1 datastart npopulation;
				run;
				data arlequin; set arlequin;
					if flag = 1 then delete;
					drop flag;
				run; 
				%NObsVars(arlequin);
				data _null_; set arlequin;
					poplen = 0;
					poplen = max(poplen, length(pop));
					if _n_ = &NObs then call symput('Poplen', poplen);
				run;
				proc sort data = arlequin;
					by hID;
				run;
				data arlequin; merge arlequin haplist;
					by hID;
					length Population $ &PopLen;
					Population = pop;
					drop pop hID;
				run;
				proc sort data=arlequin; 
					by Population;
				run;	
			%end;
		*No external haplotype list;
		%if &HapDefRow = 0 %then 
			%do;	
				data arlequin;
				infile "&Datafile" firstobs=%eval(&DataStartRow+1) TRUNCOVER EXPANDTABS lrecl=32767; 
					retain flag 1;
					retain datastart 0;
		 			retain npopulation 1;
					retain pop;
					length v1 $ 2048;
					input v1 1-2048;		
					if index(v1, 'SampleName') ^= 0 then 
					do;
						pop = substr(v1, index(v1, 'SampleName')+11);
						pop = translate(pop, '', '"', '', "'");	
					end;  			
					if index(v1, 'SampleData') ^= 0 then datastart = 1;
					if datastart = 1 & index(v1, 'SampleData') = 0 then
						do;
							flag = 0;
							freq = scan(v1, 2, " ,;&Separator") + 0;
							%do j = 1 %to &NLoci;
								Mk&j = trim(left(scan(v1, %eval(&j+2), " ,;&Separator")));		
								Mk&j = translate(Mk&j, &MissingAllele, "&MissingLabel");	
							%end;
						end;
					if trim(left(v1)) = '' then flag = 1;
					if index(v1, '}') ^= 0 then 
						do;
							datastart = 0;
							npopulation = npopulation + 1;
							flag = 1;
						end;
					if npopulation > &NPops then stop;
					drop v1 datastart npopulation;
				run;	
				data arlequin; set arlequin;
					if flag = 1 then delete;
					drop flag;
				run;
				%NObsVars(arlequin);
				data arlequin; set arlequin;
					retain poplen 0 mlen1-mlen&NLoci 0; 
					poplen = max(poplen, length(pop));
					%do i = 1 %to &NLoci;
						mlen&i = max(mlen&i, length(Mk&i));
					%end;	
					if _n_ = &NObs then
						do;
							call symput("PopLen", poplen);
							%do i = 1 %to &NLoci;
								call symput("MLen&i", mlen&i);
							%end;	
						end;
					drop poplen mlen1-mlen&NLoci;
				run;
				data arlequin; set arlequin;
					length Population $ &PopLen;
					%do i = 1 %to &NLoci;
						length A&i $ &&MLen&i;
					%end;
					Population = pop;
					%do i = 1 %to &NLoci;
						A&i = trim(left(Mk&i));
					%end;	
					drop pop Mk1-Mk&NLoci;
				run;
			%end; 
	%end;

*Genotypic data;
%if &IsHaploid = 0 %then 
	%do;
		%let N2 = %eval(&NLoci*2);
		*external haplotype list;
		%if &HapDefRow ^= 0 %then
			%do;
				/* data arlequin;
					infile "&Datafile" firstobs=%eval(&DataStartRow+1) TRUNCOVER EXPANDTABS lrecl=32767; 
					retain flag 1;
					retain datastart samplesize nind 0;
		 			retain npopulation 1;
					retain pop;
					length hID $ &Len1;
					length v1 $ 1024;
					input v1 1-1024;			
					if index(v1, 'SampleName') ^= 0 then 
						do;
							pop = substr(v1, index(v1, 'SampleName')+11);
							pop = translate(pop, '', '"', '', "'");	
							nind = 0;	
						end;  			
					put pop;
					if index(v1, 'SampleData') ^= 0 then datastart = 1;
					if datastart = 1 & index(v1, 'SampleData') = 0 then
						do;
						end;
					if trim(left(v1)) = '' then flag = 1;
					if index(v1, '}') ^= 0 then 
						do;
							datastart = 0;
							npopulation = npopulation + 1;
							flag = 1;
						end;
					if npopulation > &NPops then stop;
					drop v1 datastart npopulation; */
			%end;
		%else
			%do;
				data arlequin;
					infile "&Datafile" firstobs=%eval(&DataStartRow+1) TRUNCOVER EXPANDTABS lrecl=32767; 
					retain flag 1;
					retain datastart 0;
		 			retain npopulation 1;
					retain pop;
					length v1 $ 1024;
					input v1 1-1024;		
					if index(v1, 'SampleName') ^= 0 then 
					do;
						pop = substr(v1, index(v1, 'SampleName')+11);
						pop = translate(pop, '', '"', '', "'");	
					end;  			
					if index(v1, 'SampleData') ^= 0 then datastart = 1;
					if datastart = 1 & index(v1, 'SampleData') = 0 then
						do;
							flag = 0;
							if trim(left(scan(v1, %eval(&Nloci+1), " ,;&Separator"))) ^= '' then
								do;
									freq = scan(v1, 2, " ,;&Separator") + 0;
									%do j = 1 %to &NLoci;
										Mk&j = trim(left(scan(v1, %eval(&j+2), " ,;&Separator")));		
										Mk&j = translate(Mk&j, &MissingAllele, "&MissingLabel");	
									%end;
								end;
							else
								do;
									%do j = 1 %to &NLoci;
										%let k = %eval(&NLoci + &j);
										Mk&k = trim(left(scan(v1, &j, " ,;&Separator")));	
										Mk&j = translate(Mk&j, &MissingAllele, "&MissingLabel");		
									%end;
								end;	
						end;	
					if trim(left(v1)) = '' then flag = 1;
					if index(v1, '}') ^= 0 then 
						do;
							datastart = 0;
							npopulation = npopulation + 1;
							flag = 1;
						end;
					if npopulation > &NPops then stop;
					drop v1 datastart npopulation;
				run;
			%end;	
		data arlequin; set arlequin;
			if flag = 1 then delete;
			drop flag;
		run;
		%let NN = %eval(&NLoci+1);
		data part1; set arlequin;
			if freq ^= .;
			drop Mk&NN-Mk&N2;
		run;
		data part2; set arlequin;	
			if freq = .;
			drop Mk1-Mk&NLoci freq;
		run;
		data arlequin; merge part1 part2;
		run;
		%NObsVars(arlequin);
		data arlequin; set arlequin;
			retain poplen 0 mlen1-mlen&N2 0; 
			poplen = max(poplen, length(pop));
			%do i = 1 %to &N2;
				mlen&i = max(mlen&i, length(Mk&i));
			%end;	
			if _n_ = &NObs then
				do;
					call symput("PopLen", poplen);
					%do i = 1 %to &N2;
						call symput("MLen&i", mlen&i);
					%end;	
				end;
			drop poplen mlen1-mlen&N2;
		run;

		data arlequin; set arlequin;
			length Population $ &PopLen;
			%do i = 1 %to &NLoci;
				%let j = %eval(2*&i - 1);
				%let k = %eval(2*&i);
				%let l = %eval(&Nloci + &i);
				length A&j $ &&MLen&i A&k $&&MLen&l;
			%end;
			Population = pop;
			%do i = 1 %to &NLoci;
				%let j = %eval(2*&i - 1);
				%let k = %eval(2*&i);
				%let l = %eval(&Nloci + &i);
				A&j = trim(left(Mk&i));
				A&k = trim(left(Mk&l));
			%end;	
			drop pop Mk1-Mk&N2;
		run;
	%end;

data outlib.&OutData; set arlequin;
	do i = 1 to freq;	
		output;
	end;
	drop freq i;
run;

%let day2 = %sysfunc(round(%sysevalf(%sysfunc(date()))));
%let VeryEnd = %sysfunc(round(%sysevalf(%sysfunc(time())),.01));
%if &day1 = &day2 %then %do;
   %let TotalTime = %sysfunc(round(%sysevalf(&VeryEnd - &VeryStart),.01));
%end;
%else %do;
   %let TotalTime = %sysevalf((&day2-&day1)*86400) + %sysfunc(round(%sysevalf(&VeryEnd - &VeryStart),.01));
%end;  
data _null_;
   t = &TotalTime;
   put "The total importing time is" t time11.2;
run;
%exit:;

* Packaging;
%PackageBegin(&ProcessName);
%PackageInsertDataset(data from &ProcessName, OutLib, &OutData);
%if %upcase(&UploadInFile)=YES %then %do;
  %PackageInsertFile(Raw Data,&TmpPath,&TmpName2,,,No);
%end;
%PackagePublish(&OutPath,to_requester);
%PackageEnd;

%if (&syserr | &exiterror) %then %put ERROR: &ProcessName exited due to errors.;
%if (&clockrunning) %then %ClockStop(&ProcessName);
%mend ParseArlequin;
%ParseArlequin;