/*
SAS Genetic Marker Analytical Process

Name: Marker Properties
Author: Wendy Czika

Description:
Marker Properties calculates various single marker 
measures, such as marker informativeness, 
test for HWE, and allele and genotype frequencies.
A cell plot representing the data set containing
individuals' genotypes is created, with red
representing the genotype homozygous for the major 
allele, gray representing a heterozygous genotype,
and blue representing the genotype homozygous for
the minor allele.

See the ALLELE procedure in the SAS/Genetics User's
Guide for more information.
*/

/*
-----------------------------------------------------------
Copyright (C) 2004 SAS Institute, Inc. All rights reserved. 

Notice: 
The above copyright notice and this notice must appear in
any whole or partial copy of this code and any related
documentation. Permission to use, copy, and modify the
source code contained in this file is hereby granted, but
is limited to customers of SAS with a valid license for the
SAS software product with which this file was provided. 
Except as expressly set forth herein, the terms of such
license apply.
-----------------------------------------------------------
*/

*ProcessVariables;

%let InData = ; 
* Full path and name to the input SAS data set. 
The name must be lowercase.;

%let AlleleVars = ; 
* Variables containing marker alleles (2 per marker).;

%let ByVars = ; 
* BY variables for PROC ALLELE.;

%let WhereStmt = ;
* Optional WHERE statement for PROC ALLELE.;

%let AnnoInData = ;
* Annotation SAS data set containing marker information.;

%let AnnoLabel = ; 
* Annotation variable to be used as label for markers 
in output.;

%let Perms = 0;
* Number of permutations to use for estimate of exact p-values for HWE 
(default is 0).;

%let Boot = 0;
* Number of bootstrap samples to use for allele frequency and HW disequilibrium
confidence intervals.;

%let Alpha = 0.05;
* When Boot > 0, (1-Alpha)% confidence intervals are created.;

%let Seed = 12345;
* Random number seed;

%let ChartFreqs = no;
* Specify Yes to create frequency chart for alleles and genotypes.;

%let TransposeCellPlotData = no;
* Specify Yes to transpose the marker genotype data for the cell plot
in JMP, for example when there are more markers than there are subjects
in the input data set.;

%let OutPath = ; 
* Path for output on your SAS server. If you leave this blank the system 
will use a default location.;

%let MacroPath = ; 
* Path for SAS macros.;

%let KeepProcessInfo = No;
* Option to keep process information.  If specify "Yes", a data set, 
_processInfo, will be add to result tree;
 
*ProcessBody;

%macro MarkerProperties;

*initialization and argument checking;
%let ProcessName = markerproperties; 
%global exiterror;
%let exiterror = 0;
%let clockrunning = 0;
options nospool nolabel;
data _null_;
   infile "&MacroPath.UtilityMacros.sas";
run;
%if &syserr %then %do;
   %put ERROR: Path for SAS macros is invalid.;
   %goto exit;
%end;
%include "&MacroPath.UtilityMacros.sas" / nosource;
%ClockStart(&ProcessName);
%if &syserr %then %goto exit;
%let clockrunning = 1;

%PathName(&InData);
%let InData = &TmpName;
libname InLib "&TmpPath";

%CheckForBlank(&OutPath,Output path);
libname OutLib "&OutPath";

%if %length(&AnnoInData) %then %do;
 %PathName(&AnnoInData);
 %if &exiterror %then %do;
  	%put ERROR: Annotation data set specification is invalid.;
   	%goto exit;
 %end;
 %global AnnoData;
 %let AnnoData = &TmpName;
 libname AnnoLib "&TmpPath";

 data map;
  set AnnoLib.&AnnoData;
  name=&AnnoLabel;
 run;
%end;
%else %do;
 %let i=1;
 %do %while (%length(%scan(&AlleleVars,&i,' '))>0);
  %let i=%eval(&i+1);
 %end;
 data map;
  do j=1 to %eval(&i/2);
   name="Marker"||left(j);
   output;
  end;
 run;
%end;

%if %length(&ByVars) %then %do;
   proc sort data=InLib.&InData out=a;
      by &ByVars;
   run;
   %let ds = a;
   %let ByVars=%upcase(&ByVars);
   %let temp=%sysfunc(tranwrd(&ByVars,DESCENDING, ));
   %let ByVarsOnly=%sysfunc(compbl(&temp));
   %let JBy=%sysfunc(translate(&ByVarsOnly,","," "));  
%end;
%else %do;
 %let ds = InLib.&InData;
 %let ByVarsOnly = ;
%end;

* setup for html output;
ods html 
   body = "&ProcessName.body.html"
   frame = "&ProcessName.frame.html"   
   contents = "&ProcessName.contents.html"
   path = "&OutPath" (url=none)
;
 
ods output MarkerSumm=ms;
ods output AlleleFreq=allelefreqs;
ods output GenotypeFreq=genotypefreqs;
proc allele data=&ds ndata=map alpha=&Alpha seed=&Seed exact=&Perms boot=&Boot; 
   %if %length(&ByVars) %then %do; 
      by &ByVars;
   %end;
   var &AlleleVars;  
   %if %length(&WhereStmt) %then %do;
    where &WhereStmt;
   %end;
run;

ods html close; 

%if %upcase(&ChartFreqs)=YES %then %do;
 data allelefreqs;
  set allelefreqs;
  sortorder=_n_;
 run;

 data genotypefreqs;
  set genotypefreqs;
  sortorder=_n_;
 run;

 proc sort data=allelefreqs;
  by &ByVars Locus;
 run;

 proc sort data=genotypefreqs;
  by &ByVars Locus;
 run;

 proc sort data=ms;
  by &ByVars Locus;
 run;

 data allelefreqs2;
  merge allelefreqs ms;
  by &ByVars Locus;
  Count=2*NIndiv*Freq;
  keep &ByVars Locus Allele Count sortorder;
 run;
 
 proc sort data=allelefreqs2 out=OutLib.allelefreqs(drop=sortorder);
  by sortorder;
 run;
 
 data genotypefreqs2;
  merge genotypefreqs ms;
  by &ByVars Locus;
  Count=NIndiv*Freq;
  keep &ByVars Locus Genotype Count sortorder;
 run;
 
 proc sort data=genotypefreqs2 out=OutLib.genotypefreqs(drop=sortorder);
  by sortorder;
 run;
%end;
%else %do;
 data OutLib.allelefreqs;
  set allelefreqs;
 run;
 
 data OutLib.genotypefreqs;
  set genotypefreqs;
 run;
%end;

  
%if (&syserr>4) %then %goto exit;

ods output AlleleFreq=af;
proc allele data=&ds ndata=map; 
   var &AlleleVars;  
   %if %length(&WhereStmt) %then %do;
    where &WhereStmt;
   %end;
run;

data af;
 set af;
 by Locus notsorted;
 if first.Locus then order+1;
run;

proc sort data=af;
 by order descending freq;
run;

data afsub;
 set af;
 by order;
 if first.order;
run;

data _null_;
 set afsub;
 call symput('MajA'||left(_n_), compress(Allele));
 call symput('Label'||left(_n_), compress(Locus));
run;

data count; 
 set &ds;
 keep &AlleleVars;
run;

%let dsid=%sysfunc(open(count));
%let n2snps=%sysfunc(attrn(&dsid,NVARS));
%let nobs=%sysfunc(attrn(&dsid,NOBS));
%let ret=%sysfunc(close(&dsid));
%let nm=%eval(&n2snps/2);

data OutLib.markergenotypes;
 set &ds;
 %if %length(&WhereStmt) %then %do;
  if &WhereStmt;
 %end;
 array a_{&n2snps} &AlleleVars;
 %do i=1 %to &nm;
  j1=%eval(2*&i-1);
  j2=%eval(2*&i);
  if a_{j1}="&&MajA&i" then do;
   if a_{j2}="&&MajA&i" then do;
    g&i=2;
   end;
   else do;
    g&i=1;
   end;
  end;
  else do; 
   if a_{j2}="&&MajA&i" then do;
    g&i=1;
   end;
  else g&i=0;
  end;
  label g&i="&&Label&i";
 %end;
 keep &ByVarsOnly g1-g&nm;
 run;
 
%if (%upcase(&TransposeCellPlotData)=YES) %then %do;
 proc transpose data=OutLib.markergenotypes out=OutLib.markergenotypes name=Marker prefix=Subject;
  var g1-g&nm;
 run;
%end;

* create JSL file;
%let JSLFileName = &ProcessName..jsl;
filename jslfile "&OutPath.&JSLFileName";

* put JSL statements;
data _null_;
   file jslfile;
   put 'data=open(".\markergenotypes.sas7bdat");';
   put 'cp=New Window( "Cell Plot of Marker Genotypes",';
   put 'Cell Plot(Y(';
   %if (%upcase(&TransposeCellPlotData)=YES) %then %do;
    %do i=1 %to &nobs-1;
     put "%str(:Column(%"Subject&i%"),)";
    %end;
    put "%str(:Column(%"Subject&nobs%")%))";
	put '));';
   %end;
   %else %do;
    %do i=1 %to &nm-1;
     put "%str(:Column(%"&&Label&i%"),)";
    %end;
    put "%str(:Column(%"&&Label&nm%")%))";
    %if %length(&ByVars)>0 %then %do;
 	 put ", By(&JBy)));";
    %end;
    %else %do; 
	 put '));';
    %end;
   %end;			

   %if %upcase(&ChartFreqs)=YES %then %do;
    put 'data=open(".\allelefreqs.sas7bdat");';
    put 'af=New Window( "Allele Frequencies",';
    put 'Chart(Freq( :Count), X(Allele), Y(N), Show Y Legend(0),';
	 		put 'Show Level Legend(1), Bar Chart(1)';
			put ', By(Locus) ';
			%if %length(&ByVars)>0 %then %do;
	 		put ", Grouping(&JBy)));";
			%end;
			%else %do;
			 put ')); ';
			%end;
			
    put 'data=open(".\genotypefreqs.sas7bdat");';
    put 'cp=New Window( "Genotype Frequencies",';
    put 'Chart(Freq( :Count), X(Genotype), Y(N), Show Y Legend(0),';
			put 'Show Level Legend(1), Bar Chart(1)';
			put ', By(Locus) ';
			%if %length(&ByVars)>0 %then %do;
	 		put ", Grouping(&JBy))); ";
			%end;
			%else %do;
			 put ')); ';
			%end;
   %end;
run;
 
* Packaging;
%PackageBegin(&ProcessName);
%PackageInsertHtml(html from &ProcessName, &OutPath, &ProcessName.body.html,
   &ProcessName.frame.html, &ProcessName.contents.html);
%if %upcase(&KeepProcessInfo) = YES %then %do;
   %PackageInsertDataset(process information of &ProcessName, OutLib, _processinfo);
%end;
%PackageInsertDataset(data from &ProcessName, OutLib, markergenotypes);
%PackageInsertDataset(data from &ProcessName, OutLib, allelefreqs);
%PackageInsertDataset(data from &ProcessName, OutLib, genotypefreqs);
%PackageInsertFile(jsl code from &ProcessName, &OutPath, &JSLFileName, "markergenotypes.sas7bdat, allelefreqs.sas7bdat, genotypefreqs.sas7bdat", text/plain);
%PackagePublish(&OutPath,to_requester);
%PackageEnd;

%exit:;
%if (&syserr | &exiterror) %then %put ERROR: &ProcessName exited due to errors.;
%if (&clockrunning) %then %do;
   %ClockStop(&ProcessName)
%end;

%mend;

* run the macro;
%MarkerProperties;






















































































